/*

*/

import tree.*;
import tokenizer.*;
import treelanguage.*;
import language5.*;

/**
   This program parses and then evaluates strings from Language_5.
*/
public class Language_5_Examples
{
   public static void main(String[] args)
   {
      // IMPORTANT: Set this to 0 or 1 depending on whether you need
      // to see all of the interpreter's debugging information.
      EvaluateAST.DEBUG = 1;

      String[] programs =
      {
         // 0, demonstrate a function call
         String.join("\n",
         "(prog",
         "  (var y 6) ",
         "  (fun f (lambda x (* x y)))",  // y is a non-local reference
         "  (var z (apply f 5)))"),

         // 1, demonstrate a function with multiple parameters
         String.join("\n",
         "(prog",
         "  (var w 10)",
         "  (fun f (lambda x y z (+ x y z w)))",  // w is a non-local reference
         "  (var x (apply f 1 2 3))",
         "  (var y (apply f 4 5 x))",
         "  (var z (apply f (+ w x y) 10 20)))"),

         // 2, demonstrate composition of functions
         String.join("\n",
         "(prog",
         "  (var w 2)",
         "  (fun f (lambda x (* x x w)))",
         "  (fun g (lambda y (+ y y w)))",
         "  (var z (apply g (apply f 5))))"),  // function composition

         // 3, demonstrate caling a function from within another function
         String.join("\n",
         "(prog",
         "  (var w 2)",
         "  (fun f (lambda x (var v (* x x w))))",
         "  (fun g (lambda y (var u (+ y w (apply f y)))))",
         "  (var z (apply g 5)))"),

         // 4, demonstrate local variables
         String.join("\n",
         "(prog",
         "  (var w 10)",
         "  (fun f (lambda x (var u (* x x w))))",
         "  (begin",
         "    (var x 0)",
         "    (begin",
         "      (var y 2)",
         "      (begin",
         "        (set x w)",
         "        (var z (apply f (+ x y w)))))))"),

         // 5, demonstrate local variables inside a function body
         String.join("\n",
         "(prog",
         "  (var n 1)",
         "  (fun f (lambda x (* n x)))",
         "  (fun g (lambda x (begin",
         "                       (var n 5)",
         "                       (begin",
         "                         (var r (apply f x))))))",
         "  (var z (apply g 2)))"), // result =  2 with static scope
                                    // result = 10 with dynamic scope

         // 6, recursive function
         String.join("\n",
         "(prog",
         "  (fun fac (lambda n",
         "     (if (<= n 0)",
         "          1",
         "          (* n (apply fac (- n 1))))))",
         "  (apply fac 6))"),

         // 7, recursive function
         String.join("\n",
         "(prog",
         "  (fun even (lambda n",
         "     (&& (!= n 1)      // use short-circuiting",
         "         (|| (<= n 0)  // use short-circuiting",
         "             (apply even (- n 2))))))",
         "  (print (apply even 7))",
         "  (print (apply even 6))",
         "  (print (apply even 0))",
         "  (apply even -1))"),

         // 8, mutually recursive functions
         String.join("\n",
         "(prog",
         "  (fun even (lambda n",
         "     (|| (<= n 0 )      // use short-circuiting",
         "         (apply odd (- n 1)))))",
         "  (fun odd (lambda n",
         "     (&& (>  n 0)       // use short-circuiting",
         "         (|| (== n 1)   // use short-circuiting",
         "             (apply even (- n 1))))))",
         "  (print (apply even 8))",
         "  (print (apply odd 7))",
         "  (print (apply odd 6))",
         "  (print (apply even 3))",
         "  (apply even -1))")
      };

      int i = 0;
      for (final String str : programs)
      {
         System.out.println("\n" + i + " =========================================");

         // Build and evaluate the abstract syntax tree that represents the program.
         try
         {
            final Tree<Token> ast = Parse2AST.parse( str );

            // Print the AST as an S-expression
            //System.out.println( EvaluateAST.toString(ast) + "\n" );

            // Pretty-print the abstract syntax tree.
            //System.out.println( PrettyPrintAST.prettyPrint( ast ) + "\n" );

            System.out.println("// Unparsed program: " );
            // Print the infix version of the expression.
            System.out.println( AST2infix.ast2infix( ast ) + "\n" );

            // Create dot and png files from the AST.
            final String baseName = String.format("Language_5_Examples.%02d", i);
            Tree2dot.tree2dot(ast, baseName);
            Tree2png.tree2png(baseName);

            // Create an xml file from the AST.
            AST2xml.ast2xml(ast, baseName);

            // Run the program (interpret the AST).
            final Value value = EvaluateAST.eval( ast );
            if (EvaluateAST.DEBUG > 0)
            {
               System.out.println("result = " + value + "\n" );
            }
            else
            {
               System.out.println("result = " + value.toSimpleString() + "\n" );
            }
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
          //e.printStackTrace();
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
          //e.printStackTrace();
         }
         catch (EvalError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
          //e.printStackTrace();
         }
         ++i;
      }
   }
}
