/*

*/

import tree.*;
import tokenizer.*;
import treelanguage.*;
import language5.*;

/**
   This program parses and then evaluates strings from Language_5.

   These examples demonstrate two ways that a function can be
   used as a first class value.
   (1) A function can be passed as a parameter to another function.
   (2) A function can be returned by another function
*/
public class Language_5_Examples_2
{
   public static void main(String[] args)
   {
      // IMPORTANT: Set this to 0 or 1 depending on whether you need
      // to see all of the interpreter's debugging information.
      EvaluateAST.DEBUG = 1;

      String[] programs =
      {
         // 0. Function with a function parameter.
         String.join("\n",
         "(prog",
         "  (var y 5)",
         "  (fun f (lambda x (* x x)))",
         "  (fun g (lambda h (apply h y)))  // y is a non-local reference",
         "  (begin",
         "    (var y 7)",
         "    (var r (apply g f))))"), // returns 25 using static scope
                                       // (returns 49 using dynamic scope)

         // 1. Function with a function return value.
         String.join("\n",
         "(prog",
           "(fun f (lambda x (* x x)))",
           "(fun g (lambda x (* x x x)))",
           "(fun h (lambda x (if x f g))) // returns either f or g",
           "(var r1 (apply h true))       // returns f",
           "(var r2 (apply h false))      // returns g",
           "(var r3 (apply r1 5))         // returns 25",
           "(var r4 (apply r2 5))         // returns 125",
           "(var r5 (apply (apply h false) 5)))"), // returns 125

         // 2. Function with a function return value.
         String.join("\n",
         "(prog",
         "  (var y 2)",
         "  (fun f (lambda x (* x y)))     // y is a non-local reference",
         "  (fun g (lambda x (* x x y)))",
         "  (fun h (lambda x (if x f g)))  // returns either f or g",
         "  (begin",
         "    (var y 3)",
         "    (var k  (apply h true)) // NOTICE HOW var AND fun ARE RELATED",
         "    (var r1 (apply k 5))    // returns 10 (or 15 using dyn scope)",
         "    (set k  (apply h false))",
         "    (var r2 (apply k 5))))")   // returns 50 (or 75 using dyn scope)
      };

      int i = 0;
      for (final String str : programs)
      {
         System.out.println("\n" + i + " =========================================");

         // Build and evaluate the abstract syntax tree that represents the program.
         try
         {
            final Tree<Token> ast = Parse2AST.parse( str );

            // Print the AST as an S-expression
            //System.out.println( EvaluateAST.toString(ast) + "\n" );

            // Pretty-print the abstract syntax tree.
            //System.out.println( PrettyPrintAST.prettyPrint( ast ) + "\n" );

            System.out.println("// Unparsed program: " );
            // Print the infix version of the expression.
            System.out.println( AST2infix.ast2infix( ast ) + "\n" );

            // Create dot and png files from the AST.
            final String baseName = String.format("Language_5_Examples2.%02d", i);
            Tree2dot.tree2dot(ast, baseName);
            Tree2png.tree2png(baseName);

            // Create an xml file from the AST.
            AST2xml.ast2xml(ast, baseName);

            // Run the program (interpret the AST).
            final Value value = EvaluateAST.eval( ast );
            if (EvaluateAST.DEBUG > 0)
            {
               System.out.println("result = " + value + "\n" );
            }
            else
            {
               System.out.println("result = " + value.toSimpleString() + "\n" );
            }
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
          //e.printStackTrace();
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
          //e.printStackTrace();
         }
         catch (EvalError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
          //e.printStackTrace();
         }
         ++i;
      }
   }
}
