/*

*/

import tree.*;
import tokenizer.*;
import treelanguage.*;
import language5.*;

/**
   This program parses and then evaluates strings from Language_5.

   These examples demonstrate some unusual ways in which we can use apply.
*/
public class Language_5_Examples_3
{
   public static void main(String[] args)
   {
      // IMPORTANT: Set this to 0 or 1 depending on whether you need
      // to see all of the interpreter's debugging information.
      EvaluateAST.DEBUG = 1;

      String[] programs =
      {
         // 0.
         String.join("\n",
         "(prog",
         "  (fun f (lambda x (* x x)))",
         "  (var g f)",
         "  (var r1 (apply g 5)))"),  // returns 25

         // 1.
         String.join("\n",
         "(prog",
         "  (fun f (lambda x y (* x y)))",
         "  (fun g (lambda f))  // g has no parameters, always returns f",
         "  (apply (apply g) 5 6))"),  // returns 30

         // 2.
         String.join("\n",
         "(prog",
         "  (fun f (lambda x (* x x)))",
         "  (fun g (lambda x (* x x x)))",
         "  (var b true)",
         "  (var r1 (apply (if b f g) 5))  // returns 25",
         "  (print r1)",
         "  (set b false)",
         "  (var r2 (apply (if b f g) 5)))"),  // returns 125

         // 3.
         String.join("\n",
         "(prog",
         "  (fun f (lambda x (* x x)))",
         "  (fun g (lambda x (* x x x)))",
         "  (fun h (lambda x (if x f g)))       // returns either f or g",
         "  (var r1 (apply (apply h  true) 5))  // returns 25",
         "  (print r1)",
         "  (var r2 (apply (apply h false) 5)))"),  // returns 125

         // 4.
         String.join("\n",
         "(prog",
         "  (var y 2)",
         "  (fun f (lambda x (* x y)))    // y is a non-local reference",
         "  (fun g (lambda x (* x x y)))",
         "  (fun h (lambda x (if x f g))) // returns either f or g",
         "  (begin",
         "    (var y 3)",
         "    (var r1 (apply (apply h  true) 5))  //return 15 (dyn) or 10 (static)",
         "    (print r1)",
         "    (var r2 (apply (apply h false) 5))))"), //return 75 (dyn) or 50 (static)

         // 5.
         String.join("\n",
         "(prog",
         "  (var x 0)",
         "  (apply x 1))"),  // error, x is not a function value

         // 6.
         "(apply (&& true (== 2 (+ 1 1))) 1)", // error, not a function value

         // 7.
         "(apply (fun g (lambda x (* x x))) 5)", // error, fun is not yet an expression

         // 8.
         "(apply (lambda x (* x x)) 5)" // error, lambda is not yet an expression
      };

      int i = 0;
      for (final String str : programs)
      {
         System.out.println("\n" + i + " =========================================");

         // Build and evaluate the abstract syntax tree that represents the program.
         try
         {
            final Tree<Token> ast = Parse2AST.parse( str );

            // Print the AST as an S-expression
            //System.out.println( EvaluateAST.toString(ast) + "\n" );

            // Pretty-print the abstract syntax tree.
            //System.out.println( PrettyPrintAST.prettyPrint( ast ) + "\n" );

            System.out.println("// Unparsed program: " );
            // Print the infix version of the expression.
            System.out.println( AST2infix.ast2infix( ast ) + "\n" );

            // Create dot and png files from the AST.
            final String baseName = String.format("Language_5_Examples3.%02d", i);
            Tree2dot.tree2dot(ast, baseName);
            Tree2png.tree2png(baseName);

            // Create an xml file from the AST.
            AST2xml.ast2xml(ast, baseName);

            // Run the program (interpret the AST).
            final Value value = EvaluateAST.eval( ast );
            if (EvaluateAST.DEBUG > 0)
            {
               System.out.println("result = " + value + "\n" );
            }
            else
            {
               System.out.println("result = " + value.toSimpleString() + "\n" );
            }
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
          //e.printStackTrace();
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
          //e.printStackTrace();
         }
         catch (EvalError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
          //e.printStackTrace();
         }
         ++i;
      }
   }
}
