/*

*/

package language5;

import tree.*;
import tokenizer.*;

/**
   Convert an Abstract Syntax Tree (AST) for Language_4
   into infix notation.
   <p>
   The ast2infix() method takes an Abstract Syntax Tree
   for Language_4 and converts it into a string that
   contains the infix version of the program.
   <p>
   ast2infix() uses knowledge about operator precedence
   (in infix notation) to use the minimum number of
   parentheses.
*/
public class AST2infix
{
   /**
      This ast2infix() method is essentially
      a in-order traversal of the tree.

      @param tree  an expression {@link Tree} to be converted to infix notation
      @return the {@link String} representation of the expression {@link Tree}'s infix notation
   */
   public static String ast2infix(final Tree<Token> tree)
   {
      String result = "";

      final String node = tree.root().lexeme;

      if ( node.equals("prog") )
      {
         result += "Program\n{\n";

         // process each expression
         for (int i = 0; i < tree.subTrees().size(); ++i)
         {
            result += ast2infix( tree.subTrees().get(i), "   ", false );

            if ( tree.subTrees().get(i).root().lexeme.equals("fun") )
            {  // give function definitions an extra \n
               result += "\n";
            }

            if ( tree.subTrees().get(i).root().lexeme.equals("fun")
              || tree.subTrees().get(i).root().lexeme.equals("begin")
              || tree.subTrees().get(i).root().lexeme.equals("if")
              || tree.subTrees().get(i).root().lexeme.equals("while") )
            {  // the above expressions do not need a semicolon
               result += "\n";
            }
            else  // include a semicolon
            {
               result += ";\n";
            }
         }

         result += "}";
      }
      else  // a single expression
      {
         result += ast2infix( tree, "", false );
      }

      return result;
   }


   /**
      This ast2infix() method is essentially
      a in-order traversal of the tree.
   */
   private static String ast2infix(final Tree<Token> tree,
                                   final String indent,
                                   final boolean inline)
   {
      String result = "";

      final String node = tree.root().lexeme;

      if ( ! inline )
         result += indent;

      if ( node.equals("fun") )
      {
         // do the name of the function
         result += tree.subTrees().get(0).root().lexeme + "(";
         // get the number of formal parameters from the lambda expression
         int numberOfParameters = tree.subTrees().get(1).subTrees().size() - 1;
         // do the list of formal parameters from the lambda expression
         for (int i = 0; i < numberOfParameters; i++)
         {
            String variable = tree.subTrees().get(1).subTrees().get(i).root().lexeme;
            result += variable;
            if (i < numberOfParameters - 1)
               result += ", ";
         }
         result += ")\n";
         // do the function body from the lambda expression
         if ( tree.subTrees().get(1).subTrees().get(numberOfParameters).root().lexeme.equals("begin") )
         {
            result += ast2infix( tree.subTrees().get(1).subTrees().get(numberOfParameters), indent, false );
         }
         else // a C/C++/Java type function body
         {
            result += indent + "{\n";
            result += ast2infix( tree.subTrees().get(1).subTrees().get(numberOfParameters), indent+"   ", false );
            if ( tree.subTrees().get(1).subTrees().get(numberOfParameters).root().lexeme.equals("if")
              || tree.subTrees().get(1).subTrees().get(numberOfParameters).root().lexeme.equals("while") )
            {  // the above expressions do not need a semicolon
               result += "\n" + indent + "}";
            }
            else  // include a semicolon
               result += ";\n" + indent + "}";
         }
      }
      else if ( node.equals("apply") )
      {
          // do the function value
         result += ast2infix( tree.subTrees().get(0), indent+"   ", true) + "(";
         // do the actual parameters
         for (int i = 1; i < tree.subTrees().size(); i++)
         {
            result += ast2infix( tree.subTrees().get(i), indent+"   ", true );
            if (i < tree.subTrees().size() - 1)
               result += ", ";
         }
         result += ")";
      }
      else if ( node.equals("if") )
      {
         result += "if (" + ast2infix( tree.subTrees().get(0),
                                       indent+"       ", true ) + ")\n";

         if ( tree.subTrees().get(1).root().lexeme.equals("begin") )
         {
            result += ast2infix( tree.subTrees().get(1), indent, false );
         }
         else if ( tree.subTrees().get(1).root().lexeme.equals("while")
                || tree.subTrees().get(1).root().lexeme.equals("if") )
         {
            result += ast2infix( tree.subTrees().get(1),
                                 indent+"   ", false );
         }
         else
         {
            result += ast2infix( tree.subTrees().get(1),
                                 indent+"   ", false );
            result += ";";
         }

         result += "\n" + indent + "else\n";

         if ( tree.subTrees().get(2).root().lexeme.equals("begin") )
         {
            result += ast2infix( tree.subTrees().get(2), indent, false );
         }
         else if ( tree.subTrees().get(2).root().lexeme.equals("while")
                || tree.subTrees().get(2).root().lexeme.equals("if") )
         {
            result += ast2infix( tree.subTrees().get(2),
                                 indent+"   ", false );
         }
         else
         {
            result += ast2infix( tree.subTrees().get(2),
                                 indent+"   ", false ) + ";";
         }
      }
      else if ( node.equals("while") )
      {
         result += "while (" + ast2infix( tree.subTrees().get(0),
                                          indent+"       ", true) + ")\n";

         if ( tree.subTrees().get(1).root().lexeme.equals("begin") )
         {
            result += ast2infix( tree.subTrees().get(1), indent, false );
         }
         else if ( tree.subTrees().get(1).root().lexeme.equals("while")
                 ||tree.subTrees().get(1).root().lexeme.equals("if") )
         {
            result += ast2infix( tree.subTrees().get(1),
                                 indent+"   ", false );
         }
         else
         {
            result += ast2infix( tree.subTrees().get(1),
                                 indent+"   ", false );
            result += ";";
         }
      }
      else if ( node.equals("begin") )
      {
         result += "{\n";

         // process each expression
         for(int i = 0; i < tree.subTrees().size(); ++i)
         {
            result += ast2infix( tree.subTrees().get(i),
                                 indent + "   ", false );
            if ( tree.subTrees().get(i).root().lexeme.equals("begin")
              || tree.subTrees().get(i).root().lexeme.equals("if")
              || tree.subTrees().get(i).root().lexeme.equals("while") )
            {  // the above expressions do not need a semicolon
               result += "\n";
            }
            else  // include a semicolon
            {
               result += ";\n";
            }
         }

         result += indent + "}";
      }
      else if ( node.equals("set") )
      {
         result += "set "
                   + tree.subTrees().get(0).root().lexeme
                   + " = " + ast2infix( tree.subTrees().get(1),
                                        indent, true );
      }
      else if ( node.equals("var") )
      {
         result += "var "
                   + tree.subTrees().get(0).root().lexeme
                   + " = " + ast2infix( tree.subTrees().get(1),
                                        indent, true );
      }
      else if ( node.equals("print") )
      {
         result += "print( " + ast2infix( tree.subTrees().get(0),
                                          indent, true ) + " )";
      }
      else if ( tree.subTrees().size() == 0 )  // constant term
      {
         result += tree.root().lexeme;;
      }
      else if ( tree.subTrees().size() == 1 )  // unary operator
      {
         if ( tree.subTrees().get(0).subTrees().size() >= 2 )
         {
            result += tree.root().lexeme + "(" + ast2infix( tree.subTrees().get(0),
                                                            indent, true ) + ")";
         }
         else
         {
            result += tree.root().lexeme + ast2infix( tree.subTrees().get(0),
                                                      indent, true );
         }
      }
      else if ( tree.subTrees().size() == 2 )  // binary operator
      {
         final int middlePrecedence = precedence( tree.root().lexeme );
         final int leftPrecedence = precedence( tree.subTrees().get(0).root().lexeme );

         if ( leftPrecedence > middlePrecedence )
         {
            result += "(" + ast2infix( tree.subTrees().get(0),
                                       indent, true ) + ")";
         }
         else
         {
            result += ast2infix( tree.subTrees().get(0),
                                 indent, true );
         }

         result += " " + tree.root().lexeme + " ";

         final int rightPrecedence = precedence( tree.subTrees().get(1).root().lexeme );

         if ( rightPrecedence >= middlePrecedence )
         {
            result += "(" + ast2infix( tree.subTrees().get(1),
                                       indent, true ) + ")";
         }
         else
         {
            result += ast2infix( tree.subTrees().get(1),
                                 indent, true );
         }
      }
      else  // operators with 3 or more operands
      {
         final int operatorPrecedence = precedence( tree.root().lexeme );
         final int leftPrecedence = precedence( tree.subTrees().get(0).root().lexeme );

         if ( leftPrecedence > operatorPrecedence )
         {
            result += "(" + ast2infix( tree.subTrees().get(0),
                                       indent, true ) + ")";
         }
         else
         {
            result += ast2infix( tree.subTrees().get(0), indent, true );
         }

         for (int i = 1; i < tree.subTrees().size(); ++i)
         {
            result += " " + tree.root().lexeme + " ";

            final int rightPrecedence = precedence( tree.subTrees().get(i).root().lexeme );

            if ( rightPrecedence >= operatorPrecedence )
            {
               result += "(" + ast2infix( tree.subTrees().get(i),
                                          indent, true ) + ")";
            }
            else
            {
               result += ast2infix( tree.subTrees().get(i),
                                    indent, true );
            }
         }
      }
      return result;
   }


   // Add the 'var', 'set' and 'print' operators to this function.
   // http://www.java-tips.org/java-se-tips/java.lang/what-is-java-operator-precedence.html
   // http://www.cs.princeton.edu/introcs/11precedence/
   private static int precedence(final String op)
   {
      int result = 0;  // "highest" precedence (i.e., constants)

      if ( op.equals("!") )
      {
         result = 2;
      }
      else if ( op.equals("**") )
      {
         result = 3;
      }
      else if ( op.equals("*")
             || op.equals("/")
             || op.equals("%") )
      {
         result = 4;
      }
      else if ( op.equals("+")
             || op.equals("-") )
      {
         result = 5;
      }
      else if ( op.equals("<")
             || op.equals(">")
             || op.equals("<=")
             || op.equals(">=") )
      {
         result = 7;
      }
      else if ( op.equals("==")
             || op.equals("!=") )
      {
         result = 8;
      }
      else if ( op.equals("&&") )
      {
         result = 11;
      }
      else if ( op.equals("||") )
      {
         result = 12;
      }
      else if ( op.equals("var") )
      {
         result = 14;
      }
      else if ( op.equals("set") )
      {
         result = 14;
      }
      else if ( op.equals("print") )
      {
         result = 14;
      }

      return result;
   }
}
